# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
TR-55 Model Implementation

A mapping between variable/parameter names found in the TR-55 document
and variables used in this program are as follows:
 * `precip` is referred to as P in the report
 * `runoff` is Q
 * `evaptrans` maps to ET, the evapotranspiration
 * `inf` is the amount of water that infiltrates into the soil (in inches)
 * `init_abs` is Ia, the initial abstraction, another form of infiltration
"""

import copy

from tr55.tablelookup import lookup_cn, lookup_bmp_storage, \
    lookup_ki, is_bmp, is_built_type, make_precolumbian, \
    get_pollutants, get_bmps
from tr55.water_quality import get_volume_of_runoff, get_pollutant_load
from tr55.operations import dict_plus


def runoff_pitt(precip, land_use):
    """
    The Pitt Small Storm Hydrology method.  The output is a runoff
    value in inches.
    """
    c1 = +3.638858398e-2
    c2 = -1.243464039e-1
    c3 = +1.295682223e-1
    c4 = +9.375868043e-1
    c5 = -2.235170859e-2
    c6 = +0.170228067e+0
    c7 = -3.971810782e-1
    c8 = +3.887275538e-1
    c9 = -2.289321859e-2
    p4 = pow(precip, 4)
    p3 = pow(precip, 3)
    p2 = pow(precip, 2)

    impervious = ((c1 * p3) + (c2 * p2) + (c3 * precip) + c4) * precip
    urb_grass = ((c5 * p4) + (c6 * p3) + (c7 * p2) + (c8 * precip) + c9) * precip  # noqa

    runoff_vals = {
        'open_water':           impervious,
        'developed_low':  0.20 * impervious + 0.80 * urb_grass,
        'cluster_housing': 0.20 * impervious + 0.80 * urb_grass,
        'developed_med':  0.65 * impervious + 0.35 * urb_grass,
        'developed_high': 0.85 * impervious + 0.15 * urb_grass,
        'developed_open':     urb_grass
    }

    if land_use not in runoff_vals:
        raise Exception('Land use %s not a built-type.' % land_use)
    else:
        return min(runoff_vals[land_use], precip)


def nrcs_cutoff(precip, curve_number):
    """
    A function to find the cutoff between precipitation/curve number
    pairs that have zero runoff by definition, and those that do not.
    """
    if precip <= -1 * (2 * (curve_number - 100.0) / curve_number):
        return True
    else:
        return False


def runoff_nrcs(precip, evaptrans, soil_type, land_use):
    """
    The runoff equation from the TR-55 document.  The output is a
    runoff value in inches.
    """
    if land_use == 'cluster_housing':
        land_use = 'developed_low'
    curve_number = lookup_cn(soil_type, land_use)
    if nrcs_cutoff(precip, curve_number):
        return 0.0
    potential_retention = (1000.0 / curve_number) - 10
    initial_abs = 0.2 * potential_retention
    precip_minus_initial_abs = precip - initial_abs
    numerator = pow(precip_minus_initial_abs, 2)
    denominator = (precip_minus_initial_abs + potential_retention)
    runoff = numerator / denominator
    return min(runoff, precip - evaptrans)


def simulate_cell_day(precip, evaptrans, cell, cell_count):
    """
    Simulate a bunch of cells of the same type during a one-day event.

    `precip` is the amount of precipitation in inches.

    `evaptrans` is evapotranspiration.

    `cell` is a string which contains a soil type and land use
    separated by a colon.

    `cell_count` is the number of cells to simulate.

    The return value is a dictionary of runoff, evapotranspiration, and
    infiltration as volumes of water.
    """
    def clamp(runoff, et, inf, precip):
        """
        This function ensures that runoff + et + inf <= precip.

        NOTE: Infiltration is normally independent of the
        precipitation level, but this function introduces a slight
        dependency (that is, at very low levels of precipitation, this
        function can cause infiltration to be smaller than it
        ordinarily would be.
        """
        total = runoff + et + inf
        if (total > precip):
            scale = precip / total
            runoff *= scale
            et *= scale
            inf *= scale
        return (runoff, et, inf)

    precip = max(0.0, precip)
    soil_type, land_use, bmp = cell.lower().split(':')

    # If  there  is no  precipitation,  then  there  is no  runoff  or
    # infiltration;  however,  there  is evapotranspiration.   (It  is
    # understood that over a period of  time, this can lead to the sum
    # of the three values exceeding the total precipitation.)
    if precip == 0.0:
        return {
            'runoff-vol': 0.0,
            'et-vol': 0.0,
            'inf-vol': 0.0,
        }

    # If  the BMP  is cluster_housing  or  no_till, then  make it  the
    # land-use.  This is  done because those two types  of BMPs behave
    # more like land-uses than they do BMPs.
    if bmp and not is_bmp(bmp):
        land_use = bmp or land_use

    # When the land-use is a built-type and the level of precipitation
    # is two inches or less, use the Pitt Small Storm Hydrology Model.
    # When the land-use is a built-type but the level of precipitation
    # is higher,  the runoff is  the larger  of that predicted  by the
    # Pitt model and NRCS model.  Otherwise, return the NRCS amount.
    if is_built_type(land_use) and precip <= 2.0:
        runoff = runoff_pitt(precip, land_use)
    elif is_built_type(land_use):
        pitt_runoff = runoff_pitt(2.0, land_use)
        nrcs_runoff = runoff_nrcs(precip, evaptrans, soil_type, land_use)
        runoff = max(pitt_runoff, nrcs_runoff)
    else:
        runoff = runoff_nrcs(precip, evaptrans, soil_type, land_use)
    inf = max(0.0, precip - (evaptrans + runoff))

    (runoff, evaptrans, inf) = clamp(runoff, evaptrans, inf, precip)
    return {
        'runoff-vol': cell_count * runoff,
        'et-vol': cell_count * evaptrans,
        'inf-vol': cell_count * inf,
    }


def create_unmodified_census(census):
    """
    This creates a cell census, ignoring any modifications.  The
    output is suitable for use as input to `simulate_water_quality`.
    """
    unmod = copy.deepcopy(census)
    unmod.pop('modifications', None)
    return unmod


def create_modified_census(census):
    """
    This creates a cell census, with modifications, that is suitable
    for use as input to `simulate_water_quality`.

    For every type of cell that undergoes modification, the
    modifications are indicated with a sub-distribution under that
    cell type.
    """
    mod = copy.deepcopy(census)
    mod.pop('modifications', None)

    for (cell, subcensus) in mod['distribution'].items():
        n = subcensus['cell_count']

        changes = {
            'distribution': {
                cell: {
                    'distribution': {
                        cell: {'cell_count': n}
                    }
                }
            }
        }

        mod = dict_plus(mod, changes)

    for modification in (census.get('modifications') or []):
        for (orig_cell, subcensus) in modification['distribution'].items():
            n = subcensus['cell_count']
            soil1, land1 = orig_cell.split(':')
            soil2, land2, bmp = modification['change'].split(':')
            changed_cell = '%s:%s:%s' % (soil2 or soil1, land2 or land1, bmp)

            changes = {
                'distribution': {
                    orig_cell: {
                        'distribution': {
                            orig_cell: {'cell_count': -n},
                            changed_cell: {'cell_count': n}
                        }
                    }
                }
            }

            mod = dict_plus(mod, changes)

    return mod


def simulate_water_quality(tree, cell_res, fn,
                           pct=1.0, current_cell=None, precolumbian=False):
    """
    Perform a water quality simulation by doing simulations on each of
    the cell types (leaves), then adding them together by summing the
    values of a node's subtrees and storing them at that node.

    `tree` is the (sub)tree of cell distributions that is currently
    under consideration.

    `pct` is the percentage of calculated water volume to retain.

    `cell_res` is the size of each cell (used for turning inches of
    water into volumes of water).

    `fn` is a function that takes a cell type and a number of cells
    and returns a dictionary containing runoff, et, and inf as
    volumes.

    `current_cell` is the cell type for the present node.
    """
    # Internal node.
    if 'cell_count' in tree and 'distribution' in tree:
        n = tree['cell_count']

        # simulate subtrees
        if n != 0:
            tally = {}
            for cell, subtree in tree['distribution'].items():
                simulate_water_quality(subtree, cell_res, fn,
                                       pct, cell, precolumbian)
                subtree_ex_dist = subtree.copy()
                subtree_ex_dist.pop('distribution', None)
                tally = dict_plus(tally, subtree_ex_dist)
            tree.update(tally)  # update this node

        # effectively a leaf
        elif n == 0:
            for pol in get_pollutants():
                tree[pol] = 0.0

    # Leaf node.
    elif 'cell_count' in tree and 'distribution' not in tree:
        # the number of cells covered by this leaf
        n = tree['cell_count']

        # canonicalize the current_cell string
        split = current_cell.split(':')
        if (len(split) == 2):
            split.append('')
        if precolumbian:
            split[1] = make_precolumbian(split[1])
        current_cell = '%s:%s:%s' % tuple(split)

        # run the runoff model on this leaf
        result = fn(current_cell, n)  # runoff, et, inf
        runoff_adjustment = result['runoff-vol'] - (result['runoff-vol'] * pct)
        result['runoff-vol'] -= runoff_adjustment
        result['inf-vol'] += runoff_adjustment
        tree.update(result)

        # perform water quality calculation
        if n != 0:
            soil_type, land_use, bmp = split
            runoff_per_cell = result['runoff-vol'] / n
            liters = get_volume_of_runoff(runoff_per_cell, n, cell_res)
            for pol in get_pollutants():
                tree[pol] = get_pollutant_load(land_use, pol, liters)


def postpass(tree):
    """
    Remove volume units and replace them with inches.
    """
    if 'cell_count' in tree:
        if tree['cell_count'] > 0:
            n = tree['cell_count']
            tree['runoff'] = tree['runoff-vol'] / n
            tree['et'] = tree['et-vol'] / n
            tree['inf'] = tree['inf-vol'] / n
        else:
            tree['runoff'] = 0
            tree['et'] = 0
            tree['inf'] = 0
        tree.pop('runoff-vol', None)
        tree.pop('et-vol', None)
        tree.pop('inf-vol', None)

    if 'distribution' in tree:
        for subtree in tree['distribution'].values():
            postpass(subtree)


def compute_bmp_effect(census, m2_per_pixel):
    """
    Compute the overall percentage of pre-BMP water to retain after
    considering BMPs.
    """
    meters_per_inch = 0.0254
    cubic_meters = census['runoff-vol'] * meters_per_inch * m2_per_pixel
    bmp_dict = census.get('BMPs', {})
    bmp_keys = set(bmp_dict.keys())

    reduction = 0.0
    for bmp in set.intersection(set(get_bmps()), bmp_keys):
        bmp_area = bmp_dict[bmp]
        reduction += lookup_bmp_storage(bmp) * bmp_area
    return max(0.0, cubic_meters - reduction) / cubic_meters


def simulate_modifications(census, fn, cell_res, pc=False):
    """
    Simulate effects of modifications.

    `census` contains a distribution of cell-types in the area of interest.

    `fn` is as described in `simulate_water_quality`.

    `cell_res` is as described in `simulate_water_quality`.
    """
    mod = create_modified_census(census)
    simulate_water_quality(mod, cell_res, fn, precolumbian=pc)
    pct = compute_bmp_effect(mod, cell_res)
    simulate_water_quality(mod, cell_res, fn, pct=pct, precolumbian=pc)
    postpass(mod)

    unmod = create_unmodified_census(census)
    simulate_water_quality(unmod, cell_res, fn, precolumbian=pc)
    postpass(unmod)

    return {
        'unmodified': unmod,
        'modified': mod
    }


def simulate_day(census, precip, cell_res=10, precolumbian=False):
    """
    Simulate a day, including water quality effects of modifications.

    `census` contains a distribution of cell-types in the area of interest.

    `cell_res` is as described in `simulate_water_quality`.

    `precolumbian` indicates that artificial types should be turned
    into forest.
    """
    et_max = 0.207

    if 'modifications' in census:
        verify_census(census)

    def fn(cell, cell_count):
        # Compute et for cell type
        split = cell.split(':')
        if (len(split) == 2):
            (land_use, bmp) = split
        else:
            (_, land_use, bmp) = split
        et = et_max * lookup_ki(bmp or land_use)

        # Simulate the cell for one day
        return simulate_cell_day(precip, et, cell, cell_count)

    return simulate_modifications(census, fn, cell_res, precolumbian)


def verify_census(census):
    """
    Assures that there is no soil type/land cover pair
    in a modification census that isn't in the AoI census.
    """
    for modification in census['modifications']:
        for land_cover in modification['distribution']:
            if land_cover not in census['distribution']:
                raise ValueError("Invalid modification census")



"""
Generates a table with the model output for different input values.
This table is used in the mini-app.
"""

from itertools import product
import csv
import sys

from tr55.model import simulate_day


def cm_to_inches(cm):
    return cm / 2.54


def inches_to_cm(inches):
    return inches * 2.54


if len(sys.argv) != 2:
    print 'Usage: python -m tr55.makeMiniAppTable csv_file_name'
    sys.exit()
else:
    csv_file_name = sys.argv[1]

with open(csv_file_name, 'wb') as csv_file:
    writer = csv.writer(csv_file)
    writer.writerow(('P', 'land', 'soil', 'ET', 'I', 'R'))

    # values of inputs to the model
    precip_cm = [1, 3, 5, 8, 21]

    # The land uses in the original mini-app used non NLCD types
    # (residential, high intensity residential, commercial, etc.) These were
    # converted to the best matches from the NLCD types based on the NLCD
    # number being used for the calculations in tables.py.
    land_uses = [
        'open_water',
        'developed_open',
        'developed_low',
        'developed_med',
        'developed_high',
        'barren_land',
        'deciduous_forest',
        'shrub',
        'grassland',
        'pasture',
        'cultivated_crops',
        'woody_wetlands'
    ]

    soil_types = ['a', 'b', 'c', 'd']

    soil_type_map = {
        'a': 0,
        'b': 1,
        'c': 2,
        'd': 3
    }

    # For each input value, compute the model outputs for a
    # single day and tile.
    for precip, land_use, soil_type in product(precip_cm,
                                               land_uses,
                                               soil_types):
        cells = {
            'cell_count': 1,
            'distribution': {
                '%s:%s' % (soil_type, land_use): {'cell_count': 1}
            }
        }

        model_out = simulate_day(cells, cm_to_inches(precip))['unmodified']

        writer.writerow((precip,
                         land_use,
                         soil_type_map[soil_type],
                         inches_to_cm(model_out['et']),
                         inches_to_cm(model_out['inf']),
                         inches_to_cm(model_out['runoff'])))



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

from tr55.tablelookup import lookup_load, lookup_nlcd


def get_volume_of_runoff(runoff, cell_count, cell_resolution):
    """
    Calculate the volume of runoff over the entire modeled area

    Args:
        runoff (number): Q from TR55, averaged amount of runoff over a number
        of cells.

        cell_count (integer): The number of cells included in the area

        cell_resolution (number): The size in square meters that a cell
        represents

    Returns:
        The volume of runoff liters in of the total area of interest
    """

    # Runoff is in inches, so convert to meters which is the units for the cell
    # area and compute the meter-cells in the group.  Multiply the resolution
    # of the cell to get the runoff volume in cubic meters.
    inch_to_meter = 0.0254

    runoff_m = runoff * inch_to_meter
    meter_cells = runoff_m * cell_count
    volume_cubic_meters = meter_cells * cell_resolution

    liters = volume_cubic_meters * 1000

    return liters


def get_pollutant_load(use_type, pollutant, runoff_liters):
    """
    Calculate the pollutant load over a particular land use type given an
    amount of runoff generated on that area and an event mean concentration
    of the pollutant.  Returns the pollutant load in lbs.
    """
    mg_per_kg = 1000000
    lbs_per_kg = 2.205

    nlcd = lookup_nlcd(use_type)
    emc = lookup_load(nlcd, pollutant)

    load_mg_l = emc * runoff_liters

    return (load_mg_l / mg_per_kg) * lbs_per_kg



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
Table Lookup test set.
"""

import unittest

from tr55.tablelookup import lookup_bmp_storage, lookup_cn


class TestTablelookups(unittest.TestCase):
    """
    Table Lookup test set
    """
    def test_lookup_bmp_storage(self):
        """
        Do some spot-checks on the data from Table B.
        """
        self.assertEqual(lookup_bmp_storage('green_roof'), 0.396)
        self.assertEqual(lookup_bmp_storage('infiltration_trench'), 0.610)  # noqa
        self.assertEqual(lookup_bmp_storage('porous_paving'), 0.267)
        self.assertEqual(lookup_bmp_storage('rain_garden'), 0.396)

    def test_lookup_cn(self):
        """
        Do some spot-checks on the data from Table C.
        """
        self.assertEqual(lookup_cn('a', 'open_water'), 100)
        self.assertEqual(lookup_cn('b', 'developed_low'), 68)
        self.assertEqual(lookup_cn('c', 'developed_med'), 90)
        self.assertEqual(lookup_cn('d', 'developed_high'), 95)
        self.assertEqual(lookup_cn('a', 'barren_land'), 77)
        self.assertEqual(lookup_cn('b', 'deciduous_forest'), 55)
        self.assertEqual(lookup_cn('c', 'evergreen_forest'), 70)
        self.assertEqual(lookup_cn('d', 'mixed_forest'), 77)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
Model test set
"""

import unittest

from tr55.model import runoff_nrcs, \
    simulate_cell_day, simulate_water_quality, \
    create_unmodified_census, create_modified_census, \
    simulate_day
from tr55.tablelookup import lookup_ki

# These data are taken directly from Table 2-1 of the revised (1986)
# TR-55 report.  The data in the PS array are various precipitation
# levels, and each respective CNx array is the calculated runoff for
# that particular curve number with the given level of precipitation
# corresponding to that in PS.
PS = [1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0]  # noqa
CN55 = [0.000, 0.000, 0.000, 0.000, 0.000, 0.020, 0.080, 0.190, 0.350, 0.530, 0.740, 0.980, 1.520, 2.120, 2.780, 3.490, 4.230, 5.000, 5.790, 6.610, 7.440, 8.290]  # noqa
CN70 = [0.000, 0.030, 0.060, 0.110, 0.170, 0.240, 0.460, 0.710, 1.010, 1.330, 1.670, 2.040, 2.810, 3.620, 4.460, 5.330, 6.220, 7.130, 8.050, 8.980, 9.910, 10.85]  # noqa
CN80 = [0.080, 0.150, 0.240, 0.340, 0.440, 0.560, 0.890, 1.250, 1.640, 2.040, 2.460, 2.890, 3.780, 4.690, 5.630, 6.570, 7.520, 8.480, 9.450, 10.42, 11.39, 12.37]  # noqa
CN90 = [0.320, 0.460, 0.610, 0.760, 0.930, 1.090, 1.530, 1.980, 2.450, 2.920, 3.400, 3.880, 4.850, 5.820, 6.810, 7.790, 8.780, 9.770, 10.76, 11.76, 12.75, 13.74]  # noqa

CENSUS_1 = {
    'cell_count': 147,
    'distribution': {
        'c:developed_high': {
            'cell_count': 42
        },
        'a:deciduous_forest': {
            'cell_count': 72
        },
        'd:developed_med': {
            'cell_count': 33
        }
    },
    'modifications': [
        {
            'change': '::no_till',
            'cell_count': 30,
            'distribution': {
                'c:developed_high': {
                    'cell_count': 20
                },
                'd:developed_med': {
                    'cell_count': 10
                }
            }
        },
        {
            'change': 'd:barren_land:',
            'cell_count': 5,
            'distribution': {
                'a:deciduous_forest': {
                    'cell_count': 5
                }
            },
        }
    ]
}

DAY_OUTPUT_1 = {
    "modified": {
        "bod": 67.66870431502203,
        "cell_count": 147,
        "distribution": {
            "a:deciduous_forest": {
                "bod": 5.1614500050486845,
                "cell_count": 72,
                "distribution": {
                    "a:deciduous_forest": {
                        "bod": 0,
                        "cell_count": 67,
                        "et": 0.14489999999999997,
                        "inf": 1.8551,
                        "runoff": 0,
                        "tn": 0,
                        "tp": 0,
                        "tss": 0
                    },
                    "d:barren_land:": {
                        "bod": 5.1614500050486845,
                        "cell_count": 5,
                        "et": 0,
                        "inf": 0.6036783267219616,
                        "runoff": 1.3963216732780384,
                        "tn": 0.0003910189397764155,
                        "tp": 0.000039101893977641545,
                        "tss": 0.03910189397764155
                    }
                },
                "et": 0.13483749999999997,
                "inf": 1.7681957171334695,
                "runoff": 0.09696678286653043,
                "tn": 0.0003910189397764155,
                "tp": 0.000039101893977641545,
                "tss": 0.03910189397764155
            },
            "c:developed_high": {
                "bod": 35.51554384871639,
                "cell_count": 42,
                "distribution": {
                    "c:developed_high": {
                        "bod": 26.505034083749653,
                        "cell_count": 22,
                        "et": 0.012419999999999999,
                        "inf": 0.2528108488309999,
                        "runoff": 1.734769151169,
                        "tn": 0.1945127501307434,
                        "tp": 0.030780039581128623,
                        "tss": 5.405744451435715
                    },
                    "c:developed_high:no_till": {
                        "bod": 9.010509764966741,
                        "cell_count": 20,
                        "et": 0.1863,
                        "inf": 1.1649828634779853,
                        "runoff": 0.6487171365220146,
                        "tn": 0.06612551521064301,
                        "tp": 0.010463817791574277,
                        "tss": 1.8377079996452328
                    }
                },
                "et": 0.09522,
                "inf": 0.6871784748533739,
                "runoff": 1.217601525146626,
                "tn": 0.2606382653413864,
                "tp": 0.0412438573727029,
                "tss": 7.243452451080948
            },
            "d:developed_med": {
                "bod": 26.991710461256947,
                "cell_count": 33,
                "distribution": {
                    "d:developed_med": {
                        "bod": 21.380810693159994,
                        "cell_count": 23,
                        "et": 0.037259999999999995,
                        "inf": 0.5561354609789999,
                        "runoff": 1.4066045390209998,
                        "tn": 0.12411741800690337,
                        "tp": 0.020837230760283047,
                        "tss": 2.5548256845216613
                    },
                    "d:developed_med:no_till": {
                        "bod": 5.610899768096954,
                        "cell_count": 10,
                        "et": 0.1863,
                        "inf": 0.9646990173599737,
                        "runoff": 0.8490009826400262,
                        "tn": 0.03257174865378317,
                        "tp": 0.005468249773992795,
                        "tss": 0.6704549722895514
                    }
                },
                "et": 0.08242363636363635,
                "inf": 0.6799425992762647,
                "runoff": 1.2376337643600988,
                "tn": 0.15668916666068655,
                "tp": 0.02630548053427584,
                "tss": 3.2252806568112127
            }
        },
        "et": 0.11175183673469387,
        "inf": 1.215031927575294,
        "runoff": 0.6732162356900119,
        "tn": 0.41771845094184934,
        "tp": 0.06758843980095638,
        "tss": 10.507835001869802
    },
    "unmodified": {
        "bod": 81.27733495679115,
        "cell_count": 147,
        "distribution": {
            "a:deciduous_forest": {
                "bod": 0,
                "cell_count": 72,
                "et": 0.14489999999999997,
                "inf": 1.8550999999999997,
                "runoff": 0,
                "tn": 0,
                "tp": 0,
                "tss": 0
            },
            "c:developed_high": {
                "bod": 50.60051961443115,
                "cell_count": 42,
                "et": 0.01242,
                "inf": 0.2528108488309999,
                "runoff": 1.734769151169,
                "tn": 0.37134252297687376,
                "tp": 0.058761893745791015,
                "tss": 10.320057589104549
            },
            "d:developed_med": {
                "bod": 30.67681534236,
                "cell_count": 33,
                "et": 0.037259999999999995,
                "inf": 0.5561354609789999,
                "runoff": 1.406604539021,
                "tn": 0.17808151279251355,
                "tp": 0.029896896308232203,
                "tss": 3.665619460400644
            }
        },
        "et": 0.08288448979591835,
        "inf": 1.1056988153959795,
        "runoff": 0.8114166948081021,
        "tn": 0.5494240357693874,
        "tp": 0.08865879005402322,
        "tss": 13.985677049505192
    }
}

CENSUS_2 = {
    'cell_count': 40,
    'BMPs': {'rain_garden': 33},
    'distribution': {
        'd:developed_med': {'cell_count': 10},
        'c:developed_high': {'cell_count': 10},
        'a:deciduous_forest': {'cell_count': 10},
        'b:pasture': {'cell_count': 10}
    },
    'modifications': [
        {
            'change': '::no_till',
            'cell_count': 1,
            'distribution': {
                'b:pasture': {'cell_count': 1}
            }
        },
        {
            'change': '::cluster_housing',
            'cell_count': 1,
            'distribution': {
                'd:developed_med': {'cell_count': 1}
            }
        }
    ]
}

DAY_OUTPUT_2 = {
    'unmodified': {
        'BMPs': {
            'rain_garden': 33
        },
        'inf': 1.1166794893660754,
        'cell_count': 40,
        'tp': 0.024279302519227997,
        'tn': 0.151799090357274,
        'et': 0.07969499999999999,
        'runoff': 0.8036255106339244,
        'distribution': {
            'a:deciduous_forest': {
                'cell_count': 10,
                'tp': 0,
                'tn': 0,
                'runoff': 0,
                'et': 0.14489999999999997,
                'inf': 1.8550999999999997,
                'bod': 0,
                'tss': 0
            },
            'b:pasture': {
                'cell_count': 10,
                'tp': 0.0012287098889476472,
                'tn': 0.00942010914859863,
                'runoff': 0.0731283523456977,
                'et': 0.1242,
                'inf': 1.8026716476543023,
                'bod': 0.40956996298254905,
                'tss': 0.20478498149127453
            },
            'c:developed_high': {
                'cell_count': 10,
                'tp': 0.013990927082331196,
                'tn': 0.0884148864230652,
                'runoff': 1.7347691511690002,
                'et': 0.012419999999999999,
                'inf': 0.2528108488309999,
                'bod': 12.047742765340752,
                'tss': 2.457156568834417
            },
            'd:developed_med': {
                'cell_count': 10,
                'tp': 0.009059665547949153,
                'tn': 0.053964094785610164,
                'runoff': 1.406604539021,
                'et': 0.037259999999999995,
                'inf': 0.5561354609789999,
                'bod': 9.2960046492,
                'tss': 1.1107937758789828
            }
        },
        'bod': 21.753317377523302,
        'tss': 3.772735326204674
    },
    'modified': {
        'BMPs': {
            'rain_garden': 33
        },
        'inf': 1.9175105,
        'cell_count': 40,
        'tp': 0,
        'tn': 0,
        'et': 0.0824895,
        'runoff': 0,
        'distribution': {
            'a:deciduous_forest': {
                'inf': 1.8550999999999997,
                'cell_count': 10,
                'tp': 0,
                'tn': 0,
                'runoff': 0,
                'et': 0.14489999999999997,
                'distribution': {
                    'a:deciduous_forest': {
                        'cell_count': 10,
                        'tp': 0,
                        'tn': 0,
                        'runoff': 0,
                        'et': 0.14489999999999997,
                        'inf': 1.8550999999999997,
                        'bod': 0,
                        'tss': 0
                    }
                },
                'bod': 0,
                'tss': 0
            },
            'b:pasture': {
                'inf': 1.86959,
                'cell_count': 10,
                'tp': 0,
                'tn': 0,
                'runoff': 0,
                'et': 0.13040999999999997,
                'distribution': {
                    'b:pasture:no_till': {
                        'cell_count': 1,
                        'tp': 0,
                        'tn': 0,
                        'runoff': 0,
                        'et': 0.1863,
                        'inf': 1.8137,
                        'bod': 0,
                        'tss': 0
                    },
                    'b:pasture': {
                        'cell_count': 9,
                        'tp': 0,
                        'tn': 0,
                        'runoff': 0,
                        'et': 0.12419999999999999,
                        'inf': 1.8758000000000001,
                        'bod': 0,
                        'tss': 0
                    }
                },
                'bod': 0,
                'tss': 0
            },
            'c:developed_high': {
                'inf': 1.9875800000000001,
                'cell_count': 10,
                'tp': 0,
                'tn': 0,
                'runoff': 0,
                'et': 0.012419999999999999,
                'distribution': {
                    'c:developed_high': {
                        'cell_count': 10,
                        'tp': 0,
                        'tn': 0,
                        'runoff': 0,
                        'et': 0.012419999999999999,
                        'inf': 1.9875800000000001,
                        'bod': 0,
                        'tss': 0
                    }
                },
                'bod': 0,
                'tss': 0
            },
            'd:developed_med': {
                'inf': 1.9577719999999998,
                'cell_count': 10,
                'tp': 0,
                'tn': 0,
                'runoff': 0,
                'et': 0.042228,
                'distribution': {
                    'd:developed_med': {
                        'cell_count': 9,
                        'tp': 0,
                        'tn': 0,
                        'runoff': 0,
                        'et': 0.037259999999999995,
                        'inf': 1.9627399999999997,
                        'bod': 0,
                        'tss': 0
                    },
                    'd:developed_med:cluster_housing': {
                        'cell_count': 1,
                        'tp': 0,
                        'tn': 0,
                        'runoff': 0,
                        'et': 0.08693999999999999,
                        'inf': 1.9130600000000002,
                        'bod': 0,
                        'tss': 0
                    }
                },
                'bod': 0,
                'tss': 0
            }
        },
        'bod': 0,
        'tss': 0
    }
}


def simulate(precip, tile_string):
    land_use = tile_string.split(':')[1]
    ki = lookup_ki(land_use)
    return simulate_cell_day(precip, 0.207 * ki, tile_string, 1)


def average(l):
    return reduce(lambda x, y: x + y, l) / len(l)


class TestModel(unittest.TestCase):
    """
    Model test set.
    """
    def test_nrcs(self):
        """
        Test the implementation of the runoff equation.
        """
        # This pair has CN=55
        runoffs = [round(runoff_nrcs(precip, 0.0, 'b', 'deciduous_forest'), 2)
                   for precip in PS]
        # Low curve number and low P cause too-high runoff
        self.assertEqual(runoffs[4:], CN55[4:])

        # This pair has CN=70
        runoffs = [round(runoff_nrcs(precip, 0.0, 'c', 'deciduous_forest'), 2)
                   for precip in PS]
        self.assertEqual(runoffs[1:], CN70[1:])

        # This pair has CN=80
        runoffs = [round(runoff_nrcs(precip, 0.0, 'd', 'pasture'), 2)
                   for precip in PS]
        self.assertEqual(runoffs, CN80)

        # This pair has CN=90
        runoffs = [round(runoff_nrcs(precip, 0.0, 'c', 'developed_med'), 2)
                   for precip in PS]
        self.assertEqual(runoffs, CN90)

    def test_simulate_day(self):
        """
        Daily simulation.
        """
        result1 = simulate_cell_day(42, 93, 'a:barren_land:', 1)
        result2 = simulate_cell_day(42, 93, 'a:barren_land:', 2)
        self.assertEqual(result1['runoff-vol'] * 2, result2['runoff-vol'])

    def test_create_unmodified_census(self):
        """
        Test create_unmodified_census.
        """
        census = {
            "cell_count": 2,
            "distribution": {
                "a:barren_land": {"cell_count": 1},
                "a:open_water": {"cell_count": 1}
            },
            "modifications": [
                {
                    "change": "::cluster_housing",
                    "cell_count": 1,
                    "distribution": {
                        "a:barren_land": {"cell_count": 1}
                    }
                }
            ]
        }

        result = create_unmodified_census(census)
        census.pop("modifications", None)
        self.assertEqual(census, result)

    def test_create_modified_census_1(self):
        """
        create_modified_census from a census w/o modifications.
        """
        census = {
            "cell_count": 5,
            "distribution": {
                "a:barren_land": {"cell_count": 3},
                "a:open_water": {"cell_count": 2}
            }
        }

        expected = {
            "cell_count": 5,
            "distribution": {
                "a:barren_land": {
                    "cell_count": 3,
                    "distribution": {"a:barren_land": {"cell_count": 3}}
                },
                "a:open_water": {
                    "cell_count": 2,
                    "distribution": {"a:open_water": {"cell_count": 2}}
                }
            }
        }

        actual = create_modified_census(census)
        self.assertEqual(actual, expected)

    def test_create_modified_census_2(self):
        """
        create_modified_census from a census w/ trivial modifications.
        """
        census = {
            "cell_count": 3,
            "distribution": {
                "a:barren_land": {"cell_count": 2},
                "a:open_water": {"cell_count": 1}
            },
            "modifications": []
        }

        expected = {
            "cell_count": 3,
            "distribution": {
                "a:barren_land": {
                    "cell_count": 2,
                    "distribution": {"a:barren_land": {"cell_count": 2}}
                },
                "a:open_water": {
                    "cell_count": 1,
                    "distribution": {"a:open_water": {"cell_count": 1}}
                }
            }
        }

        actual = create_modified_census(census)
        self.assertEqual(actual, expected)

    def test_create_modified_census_3(self):
        """
        create_modified_census with non-trivial modifications.
        """
        census = {
            "cell_count": 144,
            "distribution": {
                "a:barren_land": {"cell_count": 55},
                "a:open_water": {"cell_count": 89}
            },
            "modifications": [
                {
                    "change": "::cluster_housing",
                    "cell_count": 34,
                    "distribution": {
                        "a:barren_land": {"cell_count": 34}
                    }
                }
            ]
        }

        expected = {
            "cell_count": 144,
            "distribution": {
                "a:barren_land": {
                    "cell_count": 55,
                    "distribution": {
                        "a:barren_land:cluster_housing": {"cell_count": 34},
                        "a:barren_land": {"cell_count": 21}
                    }
                },
                "a:open_water": {
                    "cell_count": 89,
                    "distribution": {
                        "a:open_water": {"cell_count": 89}
                    }
                }
            }
        }

        actual = create_modified_census(census)
        self.assertEqual(actual, expected)

    def test_create_modified_census_4(self):
        """
        create_modified_census with different types of changes.
        """
        census = {
            "distribution": {
                "a:developed_low": {
                    "cell_count": 3
                }
            },
            "cell_count": 3,
            "modifications": [
                {
                    "distribution": {
                        "a:developed_low": {
                            "cell_count": 1
                        }
                    },
                    "cell_count": 1,
                    "change": ":deciduous_forest:cluster_housing"
                },
                {
                    "distribution": {
                        "a:developed_low": {
                            "cell_count": 1
                        }
                    },
                    "cell_count": 1,
                    "change": ":deciduous_forest:"
                },
                {
                    "distribution": {
                        "a:developed_low": {
                            "cell_count": 1
                        }
                    },
                    "cell_count": 1,
                    "change": "::cluster_housing"
                },
            ]
        }

        expected = set([
            'a:deciduous_forest:',
            'a:developed_low',
            'a:deciduous_forest:cluster_housing',
            'a:developed_low:cluster_housing'])
        modified = create_modified_census(census)
        distrib = modified['distribution']['a:developed_low']['distribution']
        actual = set(distrib.keys())
        self.assertEqual(actual, expected)

    def test_simulate_water_quality_1(self):
        """
        Test the water quality simulation.
        """
        census = {
            "cell_count": 5,
            "distribution": {
                "a:barren_land": {"cell_count": 3},
                "a:open_water": {"cell_count": 2}
            }
        }

        def fn(cell, cell_count):
            return simulate_cell_day(5, 0.207, cell, cell_count)

        simulate_water_quality(census, 93, fn)
        left = census['distribution']['a:barren_land']
        right = census['distribution']['a:open_water']
        for key in set(census.keys()) - set(['distribution']):
            self.assertEqual(left[key] + right[key], census[key])

    def test_simulate_water_quality_2(self):
        """
        Test the water quality simulation in the presence of modifications.
        """
        census = {
            "cell_count": 3,
            "distribution": {
                "a:barren_land": {"cell_count": 2},
                "a:open_water": {"cell_count": 1}
            },
            "modifications": [
                {
                    "change": "d:developed_med:",
                    "cell_count": 1,
                    "distribution": {
                        "a:barren_land": {"cell_count": 1}
                    }
                }
            ]
        }

        census1 = create_modified_census(census)
        census2 = {
            "cell_count": 3,
            "distribution": {
                "a:barren_land": {"cell_count": 1},
                "d:developed_med": {"cell_count": 1},
                "a:open_water": {"cell_count": 1}
            }
        }

        def fn(cell, cell_count):
            return simulate_cell_day(5, 0.207, cell, cell_count)

        simulate_water_quality(census1, 93, fn)
        simulate_water_quality(census2, 93, fn)
        for key in set(census1.keys()) - set(['distribution']):
            self.assertEqual(census1[key], census2[key])

    def test_simulate_water_quality_precolumbian(self):
        """
        Test the water quality simulation in Pre-Columbian times.
        """
        census1 = {
            "cell_count": 8,
            "distribution": {
                "a:developed_med": {"cell_count": 1},
                "b:no_till": {"cell_count": 1},
                "c:pasture": {"cell_count": 1},
                "d:cultivated_crops": {"cell_count": 1},
                "a:open_water": {"cell_count": 1},
                "b:shrub": {"cell_count": 1},
                "c:barren_land": {"cell_count": 1},
                "d:developed_open": {"cell_count": 1}
            }
        }

        census2 = {
            "cell_count": 8,
            "distribution": {
                "a:mixed_forest": {"cell_count": 1},
                "b:mixed_forest": {"cell_count": 1},
                "c:mixed_forest": {"cell_count": 1},
                "d:mixed_forest": {"cell_count": 2},
                "a:open_water": {"cell_count": 1},
                "b:shrub": {"cell_count": 1},
                "c:barren_land": {"cell_count": 1}
            }
        }

        census3 = census2.copy()

        def fn(cell, cell_count):
            return simulate_cell_day(7, 0.107, cell, cell_count)

        simulate_water_quality(census1, 93, fn, precolumbian=True)
        simulate_water_quality(census2, 93, fn, precolumbian=True)
        simulate_water_quality(census3, 93, fn, precolumbian=False)

        for key in set(census1.keys()) - set(['distribution']):
            self.assertAlmostEqual(census1[key], census2[key])

        for key in set(census1.keys()) - set(['distribution']):
            self.assertAlmostEqual(census2[key], census3[key])

    def test_day_1(self):
        """
        Test the simulate_day function.
        """
        self.maxDiff = None

        precip = 2
        actual = simulate_day(CENSUS_1, precip)
        expected = DAY_OUTPUT_1
        self.assertEqual(actual, expected)

    def test_day_2(self):
        """
        Test the simulate_day function with lots of BMPs.
        """
        precip = 2
        actual = simulate_day(CENSUS_2, precip)
        expected = DAY_OUTPUT_2
        self.assertEqual(actual, expected)

    def test_day_with_invalid_census(self):
        """
        Test the simulate_day function with a census
        that has a modification census with a cover type
        that doesn't exist within the AoI census. This is
        invalid input. Each land cover type in a modification
        census must be represented in AoI census.
        """
        census = {
            'distribution': {
                'b:developed_med': {'cell_count': 400},
            },
            'cell_count': 400,
            'modifications': [
                {
                    'distribution': {
                        'b:developed_low': {'cell_count': 40}
                    },
                    'cell_count': 40,
                    'change': ':deciduous_forest:'
                },
            ]
        }

        precip = 3
        self.assertRaises(ValueError,
                          simulate_day, *(census, precip))

    def test_bmp_runoff(self):
        """
        Make sure that BMPs do not produce negative runoff.
        """
        census = {
            "cell_count": 1,
            "distribution": {
                "d:developed_med": {"cell_count": 1}
            },
            "modifications": [
                {
                    "change": "::green_roof",
                    "cell_count": 1,
                    "distribution": {
                        "d:developed_med": {"cell_count": 1}
                    }
                }
            ]
        }
        result = simulate_day(census, 0.984)
        self.assertTrue(result['modified']['runoff'] >= 0)

    def test_bmp_sum(self):
        """
        Make sure that runoff, evapotranspiration, and infiltration sum to
        precipitation.
        """
        census = {
            "cell_count": 1,
            "distribution": {
                "d:developed_med": {"cell_count": 1}
            },
            "modifications": [
                {
                    "change": ":developed_high:",
                    "cell_count": 1,
                    "distribution": {
                        "d:developed_med": {"cell_count": 1}
                    }
                }
            ]
        }

        precip = 0.984
        result = simulate_day(census, precip)
        runoff = result['modified']['runoff']
        et = result['modified']['et']
        inf = result['modified']['inf']
        total = runoff + et + inf
        self.assertAlmostEqual(total, precip)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from setuptools import setup, find_packages
from codecs import open
from os import path

# Get the long description from DESCRIPTION.rst
with open(path.join(path.abspath(path.dirname(__file__)),
          'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()

tests_require = ['nose >= 1.3.4']

setup(
    name='tr55',
    version='1.1.1',
    description='A Python implementation of TR-55.',
    long_description=long_description,
    url='https://github.com/azavea/tr-55',
    author='Azavea Inc.',
    license='Apache License 2.0',
    # TODO Determine Python version support
    # See also: https://github.com/azavea/tr-55/issues/16
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords='tr-55 watershed hydrology',
    packages=find_packages(exclude=['tests']),
    install_requires=[],
    extras_require={
        'dev': [],
        'test': tests_require,
    },
    test_suite='nose.collector',
    tests_require=tests_require,
)



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

import sys


def tandem_walk(op, neutral, pred, left, right):
    """
    Walk two similarly-structured dictionaries in tandem, performing
    the given operation when both halves satisfy the given predicate.
    """
    if pred(left) and pred(right):
        return op(left, right)
    elif pred(left) and right is None:
        return op(left, neutral)
    elif left is None and pred(right):
        return op(neutral, right)
    elif isinstance(left, dict) and right is None:
        return left.copy()
    elif left is None and isinstance(right, dict):
        return right.copy()
    elif isinstance(left, dict) and isinstance(right, dict):
        retval = {}
        union = set(left.keys()) | set(right.keys())
        for key in union:
            left_val = left.get(key)
            right_val = right.get(key)
            retval[key] = tandem_walk(op, neutral, pred, left_val, right_val)
        return retval


def is_number(obj):
    """
    Is obj a number?
    """
    if sys.version_info.major == 3:
        types = (int, float)
    else:
        types = (int, long, float)
    return isinstance(obj, types)


def dict_plus(left, right):
    """
    Sum of two similarly-structured dictionaries.
    """
    def plus(x, y):
        return x + y
    return tandem_walk(plus, 0, is_number, left, right)
